<?php

namespace Snaju\Inception\Http\Request\Validators;

use Snaju\Inception\Exception\InvalidArgumentException;
use Snaju\Inception\Http\Request\Validators\Rules\AlphaRule;
use Snaju\Inception\Http\Request\Validators\Rules\BailRule;
use Snaju\Inception\Http\Request\Validators\Rules\MaxLenRule;
use Snaju\Inception\Http\Request\Validators\Rules\MinLenRule;
use Snaju\Inception\Http\Request\Validators\Rules\NumericRule;
use Snaju\Inception\Http\Request\Validators\Rules\RequiredRule;
use Snaju\Inception\Http\Request\WebRequest;
use Snaju\Inception\InceptionCore;
use Snaju\Inception\Util\Loaders\Collection;
use Snaju\Inception\Http\Request\Validators\ValidatorInterface;

class Validator implements ValidatorInterface
{

    private WebRequest $request;

    private array $rules = [];

    private array $data = [];

    private array $errors = [];

    private Collection $defaultRules;

    private bool $usesRequestData = true;

    public function __construct(array $rules = [], array $data = [], bool $sendErrorResponses = true)
    {
        $this->request = InceptionCore::getRequest();
        $this->defaultRules = new Collection([
//            'alpha'         => AlphaRule::class,
//            'bail'          => BailRule::class,
//            'max'           => MaxLenRule::class,
//            'min'           => MinLenRule::class,
//            'numeric'       => NumericRule::class,
            'required'      => RequiredRule::class,
        ]);

        $this->rules = $rules;

        // Use the request data if no data was specified
        if (count($data) > 0) {
            $this->data = $data;
            $this->usesRequestData = false;
        }
            else {
                $this->data = $this->request->dangerouslyGetAll();
                $this->usesRequestData = true;
            }

        if (count($rules) > 0) {
            $this->validate($this->data, $sendErrorResponses);
        }
    }

    public function validate(array $data = [], bool $sendErrorResponses = false)
    {
        return response()->json([
//            'rules' => $fieldRules,
            'data' => $data,
            'sendErrorResponses' => $sendErrorResponses,
        ]);
//        $ruleArray = $this->expandDotSyntaxArrayKeys($fieldRules);

        $ruleArray = array_merge([], $this->rules);



        foreach ($ruleArray as $key => $value) {
            if (str_contains($key, '.')) {
//                $nestedArray = $this->
//               echo $key;
            }
                else {
                    $rules = $this->createRulesObjectFromString($value);

//                    if (in_array($key, $data)) {
//                        print_r($rules);

                        foreach ($rules as $ruleId => $ruleValue) {

                            if ($this->defaultRules->keyExists($ruleId)) {
                                // Retrieve the static class of the validator
                                $namespace = $this->defaultRules->get($ruleId);
                                $validator = new $namespace($key, $this->usesRequestData);
//                                echo $key;

                                $keyIsValid = $validator->isValid($ruleValue, $data);
                                
                                if (! $keyIsValid) {
                                    $this->addError($key, $validator->getErrorMessage());
                                }

//                                print_r($validator);
                            }
                                else {
                                    // TODO: throw an exception as say no validator with that key exists
                                }

                        }
//                    }
                }
//

//            if (in_array($key, $data)) {
//
//            }
        }












//        $this->addError('name', 'The name field is required');
//        $this->addError('name', 'Name must be at least 2 characters.');



//        return response()->json($ruleArray);


        // Grab all the data



        // Validate

        // If an id doesn't have any dot syntax, just look at the data as is.
        // But if using request data, look into the body of the request.

        // If id has dot syntax, just look at the data as is














        if ($this->hasErrors() && $sendErrorResponses) {
            $this->sendErrorResponse();
        }
    }

    public function hasErrors(): bool
    {
        $errors = false;

        if (count($this->errors) > 0) {
            $errors = true;
        }

        return $errors;
    }

    public function getErrors(): array
    {
        return $this->errors;
    }

    public function addRule(string $identifier, $rule)
    {
        if ($rule instanceof RuleInterface) {
            $this->defaultRules->add($identifier, $rule);
        }
        else {
            throw new InvalidArgumentException('The validation Rule must implement the RuleInterface');
        }
    }

    private function addError(string $key, string $error)
    {
        $this->errors[$key][] = $error;
    }

    private function sendErrorResponse()
    {
        // TODO: in the future take into consideration the Accept to check what type of response to send instead of just JSON responses
        return response()->json(['errors' => $this->getErrors()], 400);
    }

    private function expandDotSyntaxArrayKeys(array $arr): array
    {
        $target = [];

        foreach($arr as $key => $value) {
            if (is_array($value)) {
                $value = $this->expandDotSyntaxArrayKeys($value);
            }

//            $value = $this->createRulesObjectFromString($value);

            foreach(array_reverse(explode('.', $key)) as $key) {
                $value = [$key => $value];
            }

            $target = array_merge_recursive($target, $value);
        }

        return $target;
    }

    private function createRulesObjectFromString(string $rulesString): object
    {
        $target = [];
        $stringRules = explode('|', $rulesString);

        foreach ($stringRules as $stringRule) {
            $separator = explode(':', $stringRule);
            $rule = $separator[0];
            $value = isset($separator[1]) ? $separator[1] : true;

             // If the value contains comma separated values, make an array
            if (str_contains($value, ',')) {
                $value = explode(',', $value);
            }

            $target[$rule] = $value;
        }

        return (object)$target;
    }

}


// sequence / rules / steps
/***
 * If using data directly from request check all and no dot syntax 
 */

