<?php

namespace Snaju\Inception\ORM;

use DateTime;
use Doctrine\Common\Collections\Criteria;

abstract class DataModel
{

    protected $_d;

    protected static string $resourceId = '';

    protected array $immutable = [
        'id',
        'created_at',
        'updated_at',
        'deleted_at',
    ];

    public static function get($id)
    {
        $obj = ORM::$EntityManager->find(self::getClassName(), $id);
        return $obj;
    }

    public function getVal($paramName)
    {
        if (isset($this->{$paramName})) {
            return $this->{$paramName};
        }

        return null;
    }

    public function setVal($paramName, $val)
    {
        if (isset($this->{$paramName})) {
            $this->{$paramName} = $val;

            return $this->{$paramName};
        }

        return null;
    }

    public function setIfNotImmutable(string $property, $value)
    {
        if (! in_array($property, $this->immutable)) {
            $this->setVal($property, $value);

            if (isset($this->updated_at)) {
                $this->updated_at = new DateTime();
            }
        }
    }

    public function getVars()
    {
        return get_object_vars($this);
    }

    public function getIdField()
    {
        $meta = ORM::$EntityManager->getClassMetadata(self::getClassName());
        if ($meta instanceof ClassMetadata) {
            try {
                return $this->{$meta->getSingleIdentifierFieldName()};
            } catch (MappingException $e) {
            }
        }

        return null;
    }

    public
    static function getTop(Criteria $criteria)
    {
        $objs = self::find($criteria);
        if (count($objs) > 0) {
            return $objs[0];
        }

        return null;
    }

    public
    static function find(Criteria $criteria)
    {
        return ORM::$EntityManager->getRepository(self::getClassName())->matching($criteria)->toArray();
    }

    public
    static function findAndDelete(Criteria $criteria)
    {
        $elms = self::find($criteria);
        foreach ($elms as $elm) {
            ORM::$EntityManager->remove($elm);
        }
        ORM::$EntityManager->flush();
    }

    public
    static function count(Criteria $criteria)
    {
        return ORM::$EntityManager->getRepository(self::getClassName())->matching($criteria)->count();
    }

    public
    static function getRepository()
    {
        return ORM::$EntityManager->getRepository(self::getClassName());
    }

    public
    static function getAll()
    {
        if (ORM::$EntityManager->getRepository(self::getClassName()) != null) {
            return ORM::$EntityManager->getRepository(self::getClassName())->findAll();
        }

        return [];
    }

    public
    static function query($sqlQuery)
    {
        return ORM::$EntityManager->createQuery("select * from " . self::getClassName() . $sqlQuery);
    }

    public
    static function getClassName()
    {
        return get_called_class();
    }

    public
    static function getResourceId(string $separator = ' ', bool $lowercase = true): string
    {
        $class = self::getClassName();
        $id = isset($class::$resourceId) ? $class::$resourceId : '';

        // Fallback to class name if no resource id is set in the model
        if (empty($id)) {
            $parts = explode('\\', self::getClassName());
            $words = preg_split('/(?=[A-Z])/', lcfirst($parts[count($parts) - 1]));

            if (count($words) > 1) {
                $id = trim(implode($separator, $words));
            }
            else {
                $id = $words[0];
            }
        }

        // Apply lowercase if true
        if ($lowercase) {
            $id = strtolower($id);
        }

        return $id;
    }

    public
    static function currentTimeStamp()
    {
        return date("Y-m-d H:i:s", time());
    }

    public
    static function currentDateTime()
    {
        $dt = new \DateTime('now');
        return $dt;
    }

    public
    static function sum($field, Criteria $criteria = null)
    {
        $q = ORM::$EntityManager->createQueryBuilder();
        $q->from(self::getClassName(), "sq");
        $q->select("SUM(sq.$field) as sum");
        $q->addCriteria($criteria);
        $r = $q->getQuery()->getOneOrNullResult();
        if (is_array($r)) {
            return $r['sum'];
        }

        return 0.0;
    }

    public static function getQueryBuilder()
    {
        $q = ORM::$EntityManager->createQueryBuilder();
        $q->from(self::getClassName(), "qb");
        return $q;
    }

    private
    function objVars()
    {
        return array_merge(get_object_vars($this), $this->_d);
    }

    public
    function process($data = [])
    {
        $this->_d = $data;
    }

    public function persist()
    {
        ORM::$EntityManager->persist($this);
        return $this;
    }

    public function delete()
    {
        ORM::$EntityManager->remove($this);
        ORM::$EntityManager->flush($this);
        return $this;
    }

    public function flush()
    {
        ORM::$EntityManager->flush($this);
        return $this;
    }

    /**
     * Convenience method for calling persist() and flush().
     */
    public function save()
    {
        $this->persist();
        $this->flush();
        return $this;
    }

    public function toArray()
    {
        $a = [];
        foreach ($this->getVars() as $k => $v) {
            if ($k != "_d") {
                $a[$k] = $v;
            }
        }

        return $a;
    }

    public function isEqualTo(DataModel $dataModel)
    {
        foreach ($this->getVars() as $k => $v) {
            if ($dataModel->getVal($k) == null) {
                return false;
            } else if ($dataModel->getVal($k) != $v) {
                return false;
            }
        }

        return true;
    }
}

?>