<?php

namespace Snaju\Inception\Route\Loaders\Twig;

use inception\event\defaults\OnTwigLoadEvent;
use inception\InceptionCore;
use inception\util\Loader;
use Snaju\Inception\Route\Loaders\Twig\Extensions\CustomFunctions;
use Twig\Environment;
use Twig\TwigFunction;

class TemplateManager
{

    public static $templates = [];
    public static $env = null;
    public static $scope = [];

    public static function initDefaultScope($data = [], $writeover = false)
    {
        if (isset(self::$scope['default'])) {
            array_merge(self::$scope, $data);
        }

        self::$scope['default'] = $data;
    }

    // BUG: template scopes working, but data not loading from different scopes in twig templates
    public static function addData($key, $value, $writeover = false, $scopeName = "default")
    {
        $scope = [];
        if (!isset(self::$scope[$scopeName])) {
            self::$scope[$scopeName] = $scope;
        }

        if (isset(self::$scope[$scopeName])) {
            $scope = self::$scope[$scopeName];
            if (self::isPath($key)) {
                $scope = self::setArrayByPath($key, $value, $scope);
            } else {
                if (is_array($value) && isset($scope[$key]) && is_array($scope[$key])) {
                    if ($writeover) {
                        $scope[$key] = $value;
                    } else {
                        $scope[$key] = array_merge($scope[$key], $value);
                    }
                } else if (is_array($value) && !isset($scope[$key])) {
                    $scope[$key] = $value;
                } else if (is_array($value) && isset($scope[$key]) && !is_array($scope[$key])) {
                    $scope[$key] = array_merge([$scope[$key], $value]);
                } else {
                    /*
                     * Is a single key
                     * */
                    $scope[$key] = $value;
                }
            }
        }

        self::$scope[$scopeName] = $scope;
    }

    public static function getData($key, $scopeName = "default")
    {
        if (isset(self::$scope[$scopeName])) {
            $scope = self::$scope[$scopeName];
            if (self::isPath($key)) {
                return self::getArrayByPath($key, $scope);
            } else {
                if (isset($data[$key])) {
                    return $scope[$key];
                }
            }
        }

        return null;
    }

    private
    static function getArrayByPath($key, $array)
    {
        $parts = explode(".", $key);
        $currentPath = "";
        foreach ($parts as $part) {
            $currentPath .= $part . ".";
            $nextPath = str_replace($currentPath, "", $key);

            if (isset($array[$part])) {
                if (is_array($array[$part])) {
                    return self::getArrayByPath($nextPath, $array[$part]);
                } else {
                    return $array[$part];
                }
            }
        }

        return null;
    }

    private
    static function setArrayByPath($path, $val, $array)
    {
        $parts = explode(".", $path);

        $currentPath = "";
        $size = count($parts);
        $i = 0;
        foreach ($parts as $part) {
            $i++;
            $currentPath .= $part . ".";
            $nextPath = str_replace($currentPath, "", $path);

            if ($i < $size) {
                if (isset($array[$part]) && is_array($array[$part])) {
                    $array[$part] = self::setArrayByPath($nextPath, $val, $array[$part]);
                    break;
                } else {
                    $array[$part] = self::setArrayByPath($nextPath, $val, []);
                    break;
                }
            } else {
                $array[$part] = $val;
            }
        }

        return $array;
    }

    private
    static function isPath($key)
    {
        try {
            return (strpos($key, ".") !== false);
        } catch (\Exception $exception) {
            return false;
        }
    }

    /**
     * Set twig environment in template manager and loads custom extensions.
     * @param Environment $env
     */
    public static function setTwigEnv(Environment $env)
    {
        self::$env = $env;

        // Load custom extensions (globals, functions, filters, etc.)
        self::$env->addExtension(new CustomFunctions());
    }

    public static function register(Template $template)
    {
        self::$templates[$template->getId()] = $template;
    }

    public static function exist($id)
    {
        return (isset(self::$templates[$id]));
    }

    public static function render(string $templateId, array $data = [], string $scope = "default", Environment $env = null)
    {
        if ($env == null && self::$env instanceof Environment) {
            $env = self::$env;
        }

        if (self::exist($templateId)) {
            $env->render($templateId, array_merge(self::$scope[$scope], $data));
            return $env->render($templateId, array_merge(self::$scope[$scope], $data));
        }

        return "<p style='color: red;'>Invalid Template ID</p>";
    }

}